/**************************************************************************
 *
 * Copyright 2016 Samsung Electronics co., Ltd. All Rights Reserved.
 *
 * Contact: Konstantin Drabeniuk <k.drabeniuk@samsung.com>
 * Contact: Andrii Sokolenko <a.sokolenko@samsung.com>
 * Contact: Roman Marchenko <r.marchenko@samsung.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sub license, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
 * IN NO EVENT SHALL PRECISION INSIGHT AND/OR ITS SUPPLIERS BE LIABLE FOR
 * ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
**************************************************************************/

#include "tc_tbm.h"

/* The tbm_surface.h contains the public APIs. Therefore, the error values has TBM_SURFACE_ prefix,
 * but the values of the TBM_SURFACE_ prefix error types has the same vaules as TBM_ERROR_ prefix types.
 * We check the two error types in this TBMSurface testcase.
 */

class TBMSurface : public TBMBufmgr
{
protected:
	void SetUp();
	void TearDown();
public:
	uint32_t num_formats;
	tbm_format *formats;
	tbm_format format;

	int width, height;
	tbm_surface_h surface;
};

void TBMSurface::SetUp()
{
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	TBMBufmgr::SetUp();

	result = (tbm_surface_error_e)tbm_surface_query_formats(&formats, &num_formats);
	ASSERT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	ASSERT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_NONE));

	for (uint32_t i = 0; i < num_formats; i++)
		if (formats[i] == TBM_FORMAT_ARGB8888)
			format = TBM_FORMAT_ARGB8888;

	/* fail if there is no TBM_FORMAT_ARGB8888 format */
	ASSERT_TRUE(format == TBM_FORMAT_ARGB8888);

	width = 720;
	height = 1024;
	surface = tbm_surface_internal_create_with_flags(width, height, format, TBM_BO_DEFAULT);
	ASSERT_TRUE(surface != NULL);
	ASSERT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	ASSERT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

void TBMSurface::TearDown()
{
	if (surface)
		tbm_surface_internal_destroy(surface);
	ASSERT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	ASSERT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));

	if (formats)
		free(formats);

	TBMBufmgr::TearDown();
}

static int memory_for_invalid_param;
static tbm_surface_h invalid_surface = (tbm_surface_h)&memory_for_invalid_param;

TEST_F(TBMSurface, SurfaceCeateFailInvalidGeometry)
{
	tbm_surface_h surf1 = NULL;

	surf1 = tbm_surface_create(-1, 1024, TBM_FORMAT_ARGB8888);
	EXPECT_EQ(nullptr, surf1); // Expected Value: NULL
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	surf1 = tbm_surface_create(720, -1, TBM_FORMAT_ARGB8888);
	EXPECT_EQ(nullptr, surf1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	surf1 = tbm_surface_create(720, 1024, 0);
	EXPECT_EQ(nullptr, surf1);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceCeateSuccess)
{
	tbm_surface_h surf1 = NULL;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	for (uint32_t i = 0; i < num_formats; i++) {
		surf1 = tbm_surface_create(720, 1024, formats[i]);
		EXPECT_NE((tbm_surface_h)NULL, surf1);
		EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_NONE));
		EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
		result = (tbm_surface_error_e)tbm_surface_destroy(surf1);
		EXPECT_EQ(TBM_SURFACE_ERROR_NONE, result);
		EXPECT_EQ(TBM_ERROR_NONE, result);
	}
}

TEST_F(TBMSurface, SurfaceGetFormatFailInvalidInput)
{
	tbm_format fmt = tbm_surface_get_format(invalid_surface);
	EXPECT_EQ(0, fmt);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetFormatFailNull)
{
	tbm_format fmt = tbm_surface_get_format(NULL);
	EXPECT_EQ(0, fmt);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetFormatSuccess)
{
	tbm_format fmt = tbm_surface_get_format(surface);
	EXPECT_EQ(this->format, fmt);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

TEST_F(TBMSurface, SurfaceGetWidthFailInvalidInput)
{
	int width = tbm_surface_get_width(invalid_surface);
	EXPECT_EQ(width, TBM_SURFACE_ERROR_INVALID_PARAMETER);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetWidthFailNull)
{
	int width = tbm_surface_get_width(NULL);
	EXPECT_EQ(width, TBM_SURFACE_ERROR_INVALID_PARAMETER);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetWidthSuccess)
{
	int width = tbm_surface_get_width(surface);
	EXPECT_EQ(this->width, width);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

TEST_F(TBMSurface, SurfaceGetHeightFailInvalidInput)
{
	int height = tbm_surface_get_height(invalid_surface);
	EXPECT_EQ(height, TBM_SURFACE_ERROR_INVALID_PARAMETER);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetHeightFailNull)
{
	int height = tbm_surface_get_height(NULL);
	EXPECT_EQ(height, TBM_SURFACE_ERROR_INVALID_PARAMETER);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetHeightSuccess)
{
	int height = tbm_surface_get_height(surface);
	EXPECT_EQ(this->height, height);
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	EXPECT_STREQ(tbm_error_str(tbm_get_last_error()), tbm_error_str(TBM_ERROR_NONE));
}

TEST_F(TBMSurface, SurfaceGetInfoInvalidInput)
{
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	result = (tbm_surface_error_e)tbm_surface_get_info(invalid_surface, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetInfoFailNull)
{
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	result = (tbm_surface_error_e)tbm_surface_get_info(surface, NULL);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	result = (tbm_surface_error_e)tbm_surface_get_info(NULL, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	result = (tbm_surface_error_e)tbm_surface_get_info(NULL, NULL);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceGetInfoSuccess)
{
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	result = (tbm_surface_error_e)tbm_surface_get_info(surface, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(this->format, info.format);
	EXPECT_EQ(this->height, info.height);
	EXPECT_EQ(this->width, info.width);
	EXPECT_GT(info.num_planes, 0);
	EXPECT_GT(info.bpp, 0);
	EXPECT_GT(info.size, 0);
}

TEST_F(TBMSurface, SurfaceMapFailInvalidInput)
{
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	result = (tbm_surface_error_e)tbm_surface_map(invalid_surface, 0, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceMapFailNull)
{
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	result = (tbm_surface_error_e)tbm_surface_map(surface, 0, NULL);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));

	result = (tbm_surface_error_e)tbm_surface_map(NULL, 0, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceMapSuccess)
{
	tbm_surface_info_s info;
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	result = (tbm_surface_error_e)tbm_surface_map(surface, 0, &info);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_NONE));
	EXPECT_EQ(this->format, info.format);
	EXPECT_EQ(this->height, info.height);
	EXPECT_EQ(this->width, info.width);
	EXPECT_GT(info.num_planes, 0);
	EXPECT_GT(info.bpp, 0);
	EXPECT_GT(info.size, 0);

	result = (tbm_surface_error_e)tbm_surface_unmap(surface);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_NONE));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_NONE));
}

TEST_F(TBMSurface, SurfaceUnmapFailInvalidInput)
{
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;
	result = (tbm_surface_error_e)tbm_surface_unmap(invalid_surface);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

TEST_F(TBMSurface, SurfaceUnmapFailNull)
{
	tbm_surface_error_e result = TBM_SURFACE_ERROR_NONE;

	result = (tbm_surface_error_e)tbm_surface_unmap(NULL);
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_SURFACE_ERROR_INVALID_PARAMETER));
	EXPECT_STREQ(tbm_error_str(result), tbm_error_str(TBM_ERROR_INVALID_PARAMETER));
}

