/**************************************************************************

libtbm

Copyright 2018 Samsung Electronics co., Ltd. All Rights Reserved.

Contact: SooChan Lim <sc1.lim@samsung.com>, Sangjin Lee <lsj119@samsung.com>
Boram Park <boram1288.park@samsung.com>, Changyeon Lee <cyeon.lee@samsung.com>

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sub license, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice (including the
next paragraph) shall be included in all copies or substantial portions
of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.
IN NO EVENT SHALL PRECISION INSIGHT AND/OR ITS SUPPLIERS BE LIABLE FOR
ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

**************************************************************************/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "tbm_bufmgr_int.h"
#include "tbm_log.h"
#include <dlog.h>

#define TBM_PATH_LEN        1024

#define COLOR_RED "\x1b[31m"      /* for error */
#define COLOR_YELLOW "\x1b[33m"   /* for warning */
#define COLOR_GREEN "\x1b[32m"    /* for info */
#define COLOR_RESET "\x1b[0m"

#undef LOG_TAG
#define LOG_TAG "TBM"

static unsigned int dlog_enable = 1;
static unsigned int color_enable = 1;

static unsigned int assert_level = TBM_LOG_LEVEL_NONE;

static unsigned int log_lock_init;
static pthread_mutex_t log_lock;

unsigned int tbm_log_debug_level = TBM_LOG_LEVEL_INFO;

static int stdout_fd = -1;

void
tbm_log_enable_color(unsigned int enable)
{
	color_enable = enable;
}

void
tbm_log_enable_dlog(unsigned int enable)
{
	dlog_enable = enable;
}

void
tbm_log_set_debug_level(int level)
{
	tbm_log_debug_level = level;
}

void
tbm_log_set_assert_level(int level)
{
	assert_level = level;
}

void
tbm_log_set_path(const char *path)
{
	if (!path) {
		if (stdout_fd != -1) {
			fflush(stdout);
			close(STDOUT_FILENO);
			dup2(stdout_fd, STDOUT_FILENO);
			close(stdout_fd);
			stdout_fd = -1;
		}
	} else {
		char fd_name[TBM_PATH_LEN];
		int  log_fd = -1;
		FILE *log_fl;

		snprintf(fd_name, TBM_PATH_LEN, "%s", path);

		log_fl = fopen(fd_name, "a");
		if (!log_fl) {
			TBM_ERR("failed: open file(%s)\n", fd_name);
			return;
		}

		if (stdout_fd == -1) {
			fflush(stdout);
			stdout_fd = dup(STDOUT_FILENO);
			if (stdout_fd < 0) {
				TBM_ERR("dup failed: %m\n");
				fclose(log_fl);
				return;
			}
		}

		setvbuf(log_fl, NULL, _IOLBF, 512);
		log_fd = fileno(log_fl);

		close(STDOUT_FILENO);
		dup2(log_fd, STDOUT_FILENO);
		fclose(log_fl);
	}
}

static void
_tbm_log_vprint_stdout(int level, const char *fmt, va_list arg)
{
	char *lvl_str[] = {"TBM_NON", "TBM_ERR", "TBM_WRN", "TBM_INF", "TBM_DBG"};
	char *color[] = {COLOR_RESET, COLOR_RED, COLOR_YELLOW, COLOR_GREEN, COLOR_RESET};

	if (!log_lock_init) {
		log_lock_init = 1;
		pthread_mutex_init(&log_lock, NULL);
	}

	if (level > tbm_log_debug_level)
		return;

	pthread_mutex_lock(&log_lock);

	if (color_enable)
		printf("%s", color[level]);
	printf("[%s]", lvl_str[level]);
	if (color_enable)
		printf(COLOR_RESET);
	vprintf(fmt, arg);
	printf("\n");
	pthread_mutex_unlock(&log_lock);
}

void
tbm_log_print_stdout(int level, const char *fmt, ...)
{
	va_list arg;
	va_start(arg, fmt);
	_tbm_log_vprint_stdout(level, fmt, arg);
	va_end(arg);
}

static void
_tbm_log_dlog_print(int level, const char *fmt, va_list arg)
{
	log_priority dlog_prio;

	switch (level) {
	case TBM_LOG_LEVEL_ERR:
		dlog_prio = DLOG_ERROR;
		break;
	case TBM_LOG_LEVEL_WRN:
		dlog_prio = DLOG_WARN;
		break;
	case TBM_LOG_LEVEL_INFO:
		dlog_prio = DLOG_INFO;
		break;
	case TBM_LOG_LEVEL_DBG:
		dlog_prio = DLOG_DEBUG;
		break;
	default:
		return;
	}
	__dlog_vprint(LOG_ID_SYSTEM, dlog_prio, LOG_TAG, fmt, arg);
}

void
tbm_log_print(int level, const char *fmt, ...)
{
	va_list arg;

	if (!log_lock_init) {
		log_lock_init = 1;
		pthread_mutex_init(&log_lock, NULL);
	}

	if (level > tbm_log_debug_level)
		return;

	if (dlog_enable) {
		va_start(arg, fmt);
		_tbm_log_dlog_print(level, fmt, arg);
		va_end(arg);
	} else {
		va_start(arg, fmt);
		_tbm_log_vprint_stdout(level, fmt, arg);
		va_end(arg);
	}

	assert(level > assert_level);
}

